#  Drakkar-Software trading-backend
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import ccxt
import contextlib
try:
    from aiohttp_socks import ProxyConnectionError
except ImportError:
    # local mock in case aiohttp_socks is not available
    class ProxyConnectionError(Exception):
        pass

import trading_backend.errors
import trading_backend.enums
import trading_backend.constants


class Exchange:
    SPOT_ID = None
    MARGIN_ID = None
    FUTURE_ID = None
    IS_SPONSORING = False
    ORDER_ID = "12345"

    def __init__(self, exchange):
        self._exchange = exchange

        # add backend headers
        self._exchange.connector.add_headers(self.get_headers())

    def stop(self):
        self._exchange = None

    @classmethod
    def get_name(cls):
        return 'default'

    @classmethod
    def is_sponsoring(cls) -> bool:
        return cls.IS_SPONSORING

    async def initialize(self) -> str:
        default = f"{self.get_name().capitalize()} backend initialized."
        return f"{default} {await self._ensure_broker_status()}" if self.is_sponsoring() else default

    async def _ensure_broker_status(self):
        return f"Broker rebate is enabled."

    def get_headers(self) -> dict:
        return {}

    def get_orders_parameters(self, params=None) -> dict:
        if params is None:
            params = {}
        return params

    def _allow_withdrawal_right(self) -> bool:
        return trading_backend.constants.ALLOW_WITHDRAWAL_KEYS

    def _get_symbol(self):
        if self._exchange.exchange_manager.is_future:
            return "BTC/USDT:USDT"
        return "BTC/USDT"

    async def _inner_cancel_order(self):    
        await self._exchange.connector.client.cancel_order(self.ORDER_ID, symbol=self._get_symbol())


    async def _get_api_key_rights_using_order(self) -> list[trading_backend.enums.APIKeyRights]:
        rights = [trading_backend.enums.APIKeyRights.READING]
        try:
            with self.error_describer():
                await self._inner_cancel_order()
        except ccxt.AuthenticationError as err:
            if self._exchange.is_api_permission_error(err):
                # does not have trading permission: do not add trading permissions to rights
                pass
            else:
                self.raise_accurate_auth_error_if_any(err)
                raise
        except (ccxt.BadSymbol, ccxt.OperationFailed) as err:
            # should not happen
            raise trading_backend.errors.UnexpectedError(err) from err
        except ccxt.ExchangeError as err:
            self.raise_accurate_auth_error_if_any(err)
            if not self._exchange.is_api_permission_error(err):
                # goal of the check: we are in the expected "order not found" error scenario
                # => has trading permission
                rights.append(trading_backend.enums.APIKeyRights.SPOT_TRADING)
                rights.append(trading_backend.enums.APIKeyRights.MARGIN_TRADING)
                rights.append(trading_backend.enums.APIKeyRights.FUTURES_TRADING)
        return rights

    def raise_accurate_auth_error_if_any(self, err):
        if self._exchange.is_ip_whitelist_error(err):
            raise trading_backend.errors.APIKeyIPWhitelistError(err) from err
        if self._exchange.is_authentication_error(err):
            raise ccxt.AuthenticationError(err) from err

    async def _get_api_key_rights(self) -> list[trading_backend.enums.APIKeyRights]:
        # default implementation: fetch portfolio and don't check
        # todo implementation for each exchange as long as ccxt does not support it in unified api
        with self.error_describer():
            await self._exchange.connector.client.fetch_balance()
        return [
            trading_backend.enums.APIKeyRights.READING,
            trading_backend.enums.APIKeyRights.SPOT_TRADING,
            trading_backend.enums.APIKeyRights.FUTURES_TRADING,
            trading_backend.enums.APIKeyRights.MARGIN_TRADING,
        ]

    async def _ensure_api_key_rights(self):
        # raise trading_backend.errors.APIKeyPermissionsError on missing permissions
        rights = []
        try:
            rights = await self._get_api_key_rights()
        except ccxt.BaseError as err:
            try:
                import octobot_commons.logging as logging
                logging.get_logger(self.__class__.__name__).info(
                    f"Error when checking {self.__class__.__name__} api key rights: {err} ({err.__class__.__name__})"
                )
            except ImportError:
                pass
            raise err
        except (trading_backend.errors.ExchangeAuthError, trading_backend.errors.UnexpectedError):
            # forward error
            raise
        except Exception as err:
            try:
                import octobot_commons.logging as logging
                logging.get_logger(self.__class__.__name__).exception(
                    err, True, f"Error when getting {self.__class__.__name__} api key rights: {err}"
                )
            except ImportError:
                pass
            # non ccxt error: proceed to right checks and raise
        required_right = trading_backend.enums.APIKeyRights.SPOT_TRADING
        if self._exchange.exchange_manager.is_future:
            required_right = trading_backend.enums.APIKeyRights.FUTURES_TRADING
        if self._exchange.exchange_manager.is_margin:
            required_right = trading_backend.enums.APIKeyRights.MARGIN_TRADING
        if required_right not in rights:
            raise trading_backend.errors.APIKeyPermissionsError(
                f"{required_right.value} permission is required"
            )
        if not self._allow_withdrawal_right() and trading_backend.enums.APIKeyRights.WITHDRAWALS in rights:
            raise trading_backend.errors.APIKeyPermissionsError(
                f"This api key has withdrawal rights, please revoke it."
            )

    async def is_valid_account(self, always_check_key_rights=False) -> (bool, str):
        try:
            # 1. check account
            validity, message = await self._inner_is_valid_account()
            if not always_check_key_rights and not validity:
                return validity, message
            # 2. check api key right
            await self._ensure_api_key_rights()
            return validity, message
        except trading_backend.errors.ExchangeAuthError:
            # forward exception
            raise
        except ccxt.InvalidNonce as err:
            raise trading_backend.errors.TimeSyncError(err)
        except ccxt.ExchangeError as err:
            raise trading_backend.errors.ExchangeAuthError(err)
        except trading_backend.errors.InvalidIdError as err:
            try:
                import octobot_commons.logging as logging
                logging.get_logger(self.__class__.__name__).error(
                    f"{trading_backend.errors.InvalidIdError.__name__} when checking account validity: {err}"
                )
            except ImportError:
                pass
            raise trading_backend.errors.ExchangeAuthError(err)

    async def _inner_is_valid_account(self) -> (bool, str):
        # check account validity regarding exchange requirements, exchange specific
        return True, None

    def _get_id(self):
        if self._exchange.exchange_manager.is_future:
            return self.FUTURE_ID
        if self._exchange.exchange_manager.is_margin:
            return self.MARGIN_ID
        return self.SPOT_ID

    @contextlib.contextmanager
    def error_describer(self):
        """
        Note: does not change the exception class unless it's a proxy connection issue
        """
        try:
            yield
        except (ccxt.ExchangeNotAvailable, ccxt.AuthenticationError, ProxyConnectionError) as err:
            try:
                self._exchange.connector.raise_or_prefix_proxy_error_if_relevant(err, None)
            except ccxt.BaseError:
                raise
            except Exception as err:
                # wrap error into local trading_backend.errors.UnexpectedError
                raise trading_backend.errors.UnexpectedError(err) from err



#  Drakkar-Software trading-backend
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
import mock
import ccxt

import trading_backend.exchanges as exchanges
import trading_backend.errors
import trading_backend.enums
from trading_backend.exchanges.exchange import ProxyConnectionError
from tests import default_exchange


def test_get_name(default_exchange):
    assert exchanges.Exchange(default_exchange).get_name() == exchanges.Exchange.get_name()


def test_get_orders_parameters(default_exchange):
    assert exchanges.Exchange(default_exchange).get_orders_parameters({"a": 1}) == {"a": 1}


@pytest.mark.asyncio
async def test_is_valid_account(default_exchange):
    exchange = exchanges.Exchange(default_exchange)
    with mock.patch.object(exchange._exchange.connector.client, "fetch_balance",
                           mock.AsyncMock(return_value=None)) as fetch_balance_mock:
        assert await exchange.is_valid_account() == (True, None)
        fetch_balance_mock.assert_called_once()
    with mock.patch.object(exchange._exchange.connector.client, "fetch_balance",
                           mock.AsyncMock(side_effect=ccxt.AuthenticationError)) as fetch_balance_mock:
        with pytest.raises(trading_backend.errors.ExchangeAuthError):
            assert await exchange.is_valid_account() == (True, None)
        fetch_balance_mock.assert_called_once()
    with mock.patch.object(exchange._exchange.connector.client, "fetch_balance",
                           mock.AsyncMock(side_effect=KeyError)) as fetch_balance_mock:
        # non ccxt error: proceed to right checks and raise
        with pytest.raises(trading_backend.errors.APIKeyPermissionsError):
            assert await exchange.is_valid_account() == (True, None)
        fetch_balance_mock.assert_called_once()
    with mock.patch.object(exchange._exchange.connector.client, "fetch_balance",
                           mock.AsyncMock(side_effect=ProxyConnectionError)) as fetch_balance_mock:
        # non ccxt error: proceed to right checks and raise
        with pytest.raises(trading_backend.errors.UnexpectedError):
            assert await exchange.is_valid_account() == (True, None)
        fetch_balance_mock.assert_called_once()
    with mock.patch.object(exchange._exchange.connector.client, "fetch_balance",
                           mock.AsyncMock(side_effect=ccxt.InvalidNonce)) as fetch_balance_mock:
        with pytest.raises(trading_backend.errors.TimeSyncError):
            assert await exchange.is_valid_account() == (True, None)
        fetch_balance_mock.assert_called_once()
    with (mock.patch.object(exchange, "_inner_is_valid_account",
                           mock.AsyncMock(side_effect=trading_backend.errors.InvalidIdError))
          as _inner_is_valid_account_mock):
        with pytest.raises(trading_backend.errors.ExchangeAuthError):
            assert await exchange.is_valid_account() == (True, None)
        _inner_is_valid_account_mock.assert_called_once()


@pytest.mark.asyncio
async def test_get_api_key_rights_using_order(default_exchange):
    exchange = exchanges.Exchange(default_exchange)
    read_only_rights = [
        trading_backend.enums.APIKeyRights.READING,
    ]
    trading_rights = read_only_rights + [
        trading_backend.enums.APIKeyRights.SPOT_TRADING,
        trading_backend.enums.APIKeyRights.MARGIN_TRADING,
        trading_backend.enums.APIKeyRights.FUTURES_TRADING,
    ]
    # normal cases
    with mock.patch.object(
        exchange._exchange.connector.client, "cancel_order", mock.AsyncMock()
    ) as cancel_order_mock:
        assert await exchange._get_api_key_rights_using_order() == read_only_rights
        cancel_order_mock.assert_called_once()
    with mock.patch.object(
        exchange._exchange.connector.client, "cancel_order", mock.AsyncMock(side_effect=ccxt.OrderNotFound)
    ) as cancel_order_mock:
        assert await exchange._get_api_key_rights_using_order() == trading_rights
        cancel_order_mock.assert_called_once()

    # error case 1: is_api_permission_error = True
    with mock.patch.object(
        exchange._exchange, "is_api_permission_error", mock.Mock(return_value=True)
    ) as is_api_permission_error_mock:
        with mock.patch.object(
            exchange._exchange.connector.client, "cancel_order", mock.AsyncMock(side_effect=ccxt.AuthenticationError)
        ) as cancel_order_mock:
            assert await exchange._get_api_key_rights_using_order() == read_only_rights
            cancel_order_mock.assert_called_once()
            is_api_permission_error_mock.assert_called_once()
            is_api_permission_error_mock.reset_mock()
        with mock.patch.object(
            # random ccxt error
            exchange._exchange.connector.client, "cancel_order", mock.AsyncMock(side_effect=ccxt.BadRequest)
        ) as cancel_order_mock:
            assert await exchange._get_api_key_rights_using_order() == read_only_rights
            cancel_order_mock.assert_called_once()
            is_api_permission_error_mock.assert_called_once()
            is_api_permission_error_mock.reset_mock()
        with mock.patch.object(
            exchange._exchange.connector.client, "cancel_order", mock.AsyncMock(side_effect=ccxt.BadSymbol)
        ) as cancel_order_mock:
            # unexpected error
            with pytest.raises(trading_backend.errors.UnexpectedError):
                await exchange._get_api_key_rights_using_order()
                cancel_order_mock.assert_not_called()
                is_api_permission_error_mock.assert_not_called()

        # error case 1: is_api_permission_error = False
        with mock.patch.object(
            exchange._exchange, "is_api_permission_error", mock.Mock(return_value=False)
        ) as is_api_permission_error_mock:
            # A. no accurate error
            with mock.patch.object(
                exchange, "raise_accurate_auth_error_if_any", mock.Mock()
            ) as raise_accurate_auth_error_if_any_mock:
                with mock.patch.object(
                    exchange._exchange.connector.client, "cancel_order",
                    mock.AsyncMock(side_effect=ccxt.AuthenticationError)
                ) as cancel_order_mock:
                    with pytest.raises(ccxt.AuthenticationError):
                        await exchange._get_api_key_rights_using_order()
                    raise_accurate_auth_error_if_any_mock.assert_called_once()
                    raise_accurate_auth_error_if_any_mock.reset_mock()
                    cancel_order_mock.assert_called_once()
                    is_api_permission_error_mock.assert_called_once()
                    is_api_permission_error_mock.reset_mock()
                with mock.patch.object(
                    exchange._exchange.connector.client, "cancel_order",
                    mock.AsyncMock(side_effect=ccxt.RequestTimeout)
                ) as cancel_order_mock:
                    with pytest.raises(trading_backend.errors.UnexpectedError):
                        await exchange._get_api_key_rights_using_order()
                    cancel_order_mock.assert_called_once()
                    raise_accurate_auth_error_if_any_mock.assert_not_called()
                    is_api_permission_error_mock.assert_not_called()
                with mock.patch.object(
                    exchange._exchange.connector.client, "cancel_order",
                    mock.AsyncMock(side_effect=ccxt.InsufficientFunds)
                ) as cancel_order_mock:
                    assert await exchange._get_api_key_rights_using_order() == trading_rights
                    raise_accurate_auth_error_if_any_mock.assert_called_once()
                    raise_accurate_auth_error_if_any_mock.reset_mock()
                    is_api_permission_error_mock.assert_called_once()
                    is_api_permission_error_mock.reset_mock()
                    cancel_order_mock.assert_called_once()

            # B. accurate error
            with mock.patch.object(
                exchange, "raise_accurate_auth_error_if_any", mock.Mock(
                    side_effect=trading_backend.errors.APIKeyIPWhitelistError
                    )
            ) as raise_accurate_auth_error_if_any_mock:
                with mock.patch.object(
                    exchange._exchange.connector.client, "cancel_order",
                    mock.AsyncMock(side_effect=ccxt.AuthenticationError)
                ) as cancel_order_mock:
                    with pytest.raises(trading_backend.errors.APIKeyIPWhitelistError):
                        await exchange._get_api_key_rights_using_order()
                    raise_accurate_auth_error_if_any_mock.assert_called_once()
                    raise_accurate_auth_error_if_any_mock.reset_mock()
                    cancel_order_mock.assert_called_once()
                    is_api_permission_error_mock.assert_called_once()
                    is_api_permission_error_mock.reset_mock()
                with mock.patch.object(
                    exchange._exchange.connector.client, "cancel_order",
                    mock.AsyncMock(side_effect=ccxt.RequestTimeout)
                ) as cancel_order_mock:
                    with pytest.raises(trading_backend.errors.UnexpectedError):
                        await exchange._get_api_key_rights_using_order()
                    cancel_order_mock.assert_called_once()
                    raise_accurate_auth_error_if_any_mock.assert_not_called()
                    is_api_permission_error_mock.assert_not_called()
                with mock.patch.object(
                    exchange._exchange.connector.client, "cancel_order",
                    mock.AsyncMock(side_effect=ccxt.InsufficientFunds)
                ) as cancel_order_mock:
                    with pytest.raises(trading_backend.errors.APIKeyIPWhitelistError):
                        await exchange._get_api_key_rights_using_order()
                    raise_accurate_auth_error_if_any_mock.assert_called_once()
                    raise_accurate_auth_error_if_any_mock.reset_mock()
                    is_api_permission_error_mock.assert_not_called()
                    cancel_order_mock.assert_called_once()


def test_raise_accurate_auth_error_if_any(default_exchange):
    exchange = exchanges.Exchange(default_exchange)

    with mock.patch.object(
        exchange._exchange, "is_authentication_error", mock.Mock(return_value=False)
    ), mock.patch.object(
        exchange._exchange, "is_ip_whitelist_error", mock.Mock(return_value=False)
    ):
        assert exchange.raise_accurate_auth_error_if_any(NotImplementedError("plop")) is None

    with mock.patch.object(
        exchange._exchange, "is_authentication_error", mock.Mock(return_value=True)
    ), mock.patch.object(
        exchange._exchange, "is_ip_whitelist_error", mock.Mock(return_value=False)
    ):
        with pytest.raises(ccxt.AuthenticationError, match="plop"):
            exchange.raise_accurate_auth_error_if_any(NotImplementedError("plop"))

    with mock.patch.object(
        exchange._exchange, "is_authentication_error", mock.Mock(return_value=False)
    ), mock.patch.object(
        exchange._exchange, "is_ip_whitelist_error", mock.Mock(return_value=True)
    ):
        with pytest.raises(trading_backend.APIKeyIPWhitelistError, match="plop"):
            exchange.raise_accurate_auth_error_if_any(NotImplementedError("plop"))

    with mock.patch.object(
        exchange._exchange, "is_authentication_error", mock.Mock(return_value=True)
    ), mock.patch.object(
        exchange._exchange, "is_ip_whitelist_error", mock.Mock(return_value=True)
    ):
        with pytest.raises(trading_backend.APIKeyIPWhitelistError, match="plop"):
            exchange.raise_accurate_auth_error_if_any(NotImplementedError("plop"))


@pytest.mark.asyncio
async def test_initialize(default_exchange):
    exchange = exchanges.Exchange(default_exchange)
    init_result = await exchange.initialize()
    assert exchange.get_name().capitalize() in init_result
    assert "Broker" not in init_result


@pytest.mark.asyncio
async def test_ensure_broker_status(default_exchange):
    init_result = await exchanges.Exchange(default_exchange)._ensure_broker_status()
    assert "Broker" in init_result
    assert "enabled" in init_result
